#include <cmath>
#include <cstdio>
#include <string>
#include <vector>
#include <map>

using namespace std;

#ifdef __GNUC__
typedef long double floattype;
#define FLOATFMTSTR "Lf"
#define powfun powl
#else
typedef double floattype;
#define FLOATFMTSTR "f"
#define powfun pow
#endif

static void GPL()
{printf("\n\
    This program is free software; you can redistribute it and/or modify\n\
    it under the terms of the GNU General Public License as published by\n\
    the Free Software Foundation; either version 2 of the License, or\n\
    (at your option) any later version.\n\
\n\
    This program is distributed in the hope that it will be useful,\n\
    but WITHOUT ANY WARRANTY; without even the implied warranty of\n\
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n\
    GNU General Public License for more details.\n\
\n\
    You should have received a copy of the GNU General Public License\n\
    along with this program; if not, write to the Free Software\n\
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.\n\
");
}
static string ShortHz(floattype hz, unsigned prec=3, bool force=false)
{
    const char *etu = " kMGTPE?";
    
    if(hz < 0)hz = -hz; // Shouldn't happen
    
    for(hz += 1E-7; hz >= 10000; hz /= 1000)if(etu[1])++etu;
    
    char Buf[64];
    
    sprintf(Buf, force?"%4u":"%u", (unsigned)hz);
    
    // Fractions:    
    unsigned tmp = unsigned((hz - (unsigned)hz) * powfun(10,prec));
    
    if(tmp || force)
    {
        if(force)sprintf(Buf+32,".%%0%uu",prec);else strcpy(Buf+32, ".%u");
        sprintf(strchr(Buf,0), Buf+32, tmp);
    }
    
    sprintf(Buf+58, " %cHz", *etu);
    strcat(Buf, Buf+58+(*etu==' '));
    
    return Buf;    
}

static const struct { const char *name; floattype value;
                      operator bool() const { return name; }
                    } Bases[] =
{ 
    {"1",       1},
    {"pi",      3.141592653589793238462643383279502884197169399375105820974944592},
    {"e",       2.718281828459045235360287471352662497757247093699959574966967628},
    {"sqrt(2)", 1.41421356237309504880168872420969807856967187537694807317667973},
    {0,0}
};
class how
{
    vector<string> ways;
    class empty {public: };
    struct comb {unsigned a, b; bool operator< (const comb &c) const {return a<c.a || b<c.b; }};
    map<comb, empty> powers;
public:
    void add(unsigned ind, unsigned base, unsigned exp)
    {
        comb power = {ind, (unsigned)powfun(base, exp)};
        if(powers.find(power) != powers.end())return;
        powers[power];
        
        char calc[64];
        sprintf(calc, "*%u^%u", base,exp);
        ways.push_back(Bases[ind].name + string(calc));
    }
    string explain() const
    {
        if(!ways.size())return "?";
        string result;
        for(unsigned a=0; a<ways.size(); ++a)
        {
            if(a)result += " or ";
            result += ways[a];
        }
        return result;
    }
};

static const char *a0="alienhz";
static floattype GetHz(const char *&s,const char *const opt)
{
    if(!s||!*s){fprintf(stderr, "%s: argument to `%s' is missing\n",a0,opt);exit(-1);}
    floattype base = strtod(s, (char **)&s), exp=1;
    while(*s==' '||*s=='\t')++s;
    if(*s=='M'){exp=1E6;++s;}
    else if(*s=='m'){exp=1E-3;++s;fprintf(stderr, "%s: Are you sure you didn't mean MHz (megaherz), but mHz (milliherz)?\n", a0);}
    else if(*s=='G'){exp=1E9;++s;}
    else if(*s=='k'){exp=1E3;++s;}
    else if(*s=='T'){exp=1E12;++s;}
    else if(*s=='P'){exp=1E15;++s;}
    else if(*s=='E'){exp=1E16;++s;}
    if((*s=='H'||*s=='h')&&(s[1]=='Z'||s[1]=='z'))s+=2;
    return base * exp;
}
int main(int argc, const char *const *argv)
{
    floattype minhz = 1.42E9; /* 1.42 GHz */
    floattype maxhz = 1.64E9; /* 1.64 GHz */
    
    bool explainwhy       = false;
    bool explainconstants = true;
    bool quiet            = false;
   
    bool Help=false;
    static const char VerStr[]="AlienHz version "VERSION"\nCopyright (C) 1992,2000 Bisqwit (http://iki.fi/bisqwit/)\n";
    if(argc--)a0 = *argv++;
    for(bool argterm=false; argc > 0; ++argv,--argc)
    {
        const char *s = *argv;
    reopt:
        if(!argterm && *s=='-' && s[1])
            if(s[1]=='-')
            {
                if(!s[2])argterm=true;
                else if(string(s+2)=="help")Help=true;
                else if(string(s+2)=="quiet")quiet=true, explainwhy=explainconstants=false;
                else if(string(s+2)=="explainwhy")explainwhy=true;
                else if(string(s+2)=="noexplainconstants")explainconstants=false;
                else if(string(s+2)=="min"){s=*++argv;--argc;minhz=GetHz(s,"--min");if(*s)goto reopt;}
                else if(string(s+2)=="max"){s=*++argv;--argc;maxhz=GetHz(s,"--max");if(*s)goto reopt;}
                else if(string(s+2)=="version")goto DispVers;
                else {fprintf(stderr, "%s: unrecognized option `%s'\n",a0,s);goto sghelp;}
            }
            else while(*++s)
                switch(*s)
                {
                    case '?': case 'h': Help=true; break;
                    case 'q': quiet=true, explainwhy=explainconstants=false; break;
                    case 'v': explainwhy=true; break;
                    case 'n': explainconstants=false; break;
                    case 'i': ++s; minhz=GetHz(s,"-i"); --s; break;
                    case 'x': ++s; maxhz=GetHz(s,"-x"); --s; break;
                    case 'V': DispVers:printf("%s",VerStr);GPL();return 0;
                    default:fprintf(stderr, "%s: invalid option -- %c\n",a0,*s);goto sghelp;
                }
        else
        {
            fprintf(stderr, "%s: invalid parameter `%s'\n", a0,s);sghelp:
            fprintf(stderr, "Try `%s --help' for more information.\n",a0);return -1;
        }
    }
    if(Help)
    {
        printf("%s\nDisplays the list of the most possible radio frequences, if an alien\n"
                   "civilization was trying to contact another alien civilization (like us)\n"
                   "with radio frequencies.\n"
                   "See http://setiathome.ssl.berkeley.edu/about_seti/radio_search_1.html\n"
                   "    for more information about the SETI Radio search.\n"
                   "This program calculates the most possible\n"
                   "frequencies, based on universal constants.\n"
               "\nUsage: %s [-?qhynixV]\n\n"
               "  -h, --help                Help\n"
               "  -q, --quiet               Disable all insignificant information\n"
               "  -v, --explainwhy          Describe each calculated frequency\n"
               "  -n, --noexplainconstants  Don't display and describe the universal constants\n"
               "  -i, --min                 Set minimum frequency (default: 1.42 GHz)\n"
               "  -x, --max                 Set maximum frequency (default: 1.64 GHz)\n"
               "\nExample: alienhz --min 800M --max 3G\n"
               "\n",VerStr,a0);
        return 0;
    }
    
    map<floattype, how> results;
    
    if(!quiet)
    {
       printf(
           "The most probable messaging frequencies between %s .. %s\n"
           "\n",
           ShortHz(minhz).c_str(), ShortHz(maxhz).c_str());
    }
    
    if(explainconstants)
    {
        printf(
            "Based on the following base numbers, multiplicated with\n"
            "several possible number system bases raised to 4..60 power:\n");
            
        for(unsigned a=0; Bases[a]; ++a)
            printf("%.15"FLOATFMTSTR" ", Bases[a].value);
            
        printf("\n(");
        for(unsigned a=0; Bases[a]; ++a)
        {
            if(a>0)printf(Bases[a+1]?", ":" and ");
            printf("%s", Bases[a].name);
        }
        printf(" - some universal constants)\n\n");
    }
    
    for(unsigned base=2; base<100; ++base)
    {
        floattype res;
        for(unsigned Index=0; (res=Bases[Index].value) != 0; ++Index)
        {
            map<unsigned,unsigned> exps;
            for(unsigned exp=4; exp<60; ++exp)
            {
                floattype val = res * powfun(base, exp);
                if(val > maxhz)break;
                if(val >= minhz)results[val].add(Index, base, exp);
            }
        }
    }
        
    unsigned count=0;
    for(map<floattype, how>::const_iterator i=results.begin(); i!=results.end(); ++i)
    {
        string s = ShortHz(i->first, 9, true);
        printf("%s", s.c_str());
        if(explainwhy)printf(" (%s)", i->second.explain().c_str());
        if(count%4>=3 || explainwhy)printf("\n");else printf("%*s", 19-s.size(),"");
        ++count;
    }
    if(!quiet)
    {
        printf("\nTotal: %u\n", count);
        printf("See `%s --help' for more information.\n",a0);
    }
        
    return 0;
}
